/*
 * Common methods for ethernet drivers
 */
/*
	Copyright 1999, Be Incorporated.   All Rights Reserved.
	This file may be used under the terms of the Be Sample Code License.
*/

#include "AtmDevice.h"
#include "netdebug.h"
#include "speed_touch_ctl.h"
#include <string.h>
#include <errno.h>
#include <new>
#include <BeBuild.h>



#define compat_open(dev) open(dev, O_RDWR, 0)


int
AtmDevice::Init(
				  const char *devicelink,
				  char *params
				  )
{
	T_atmVcc vcc;
	
	
	fd = compat_open(devicelink);
	if (fd < 0) {
		return (-1);
	}
	
	// TODO VPI and VCI should be taken from params
	vcc.vpi = 8;
	vcc.vci = 35;
	if (ioctl(fd, STUSB_SET_VCC, &vcc, sizeof(vcc)) < 0) {
		dprintf("stusb_set_vcc failed\n");
		close(fd);
		fd = -1;
		return (-1);
	}
	return (fd);
}

static int
toss(int fd)
{
	char junk[1536];

	return (read(fd, junk, sizeof(junk)));
}


BNetPacket *
AtmDevice::read_packet(void)
{
	BNetPacket *packet;
	int len;

	packet = NULL;
	try {
		/*
		 * Normal case: assume we don't run out of memory
		 */
		packet = AllocPacket();
		packet->SetSize(MaxPacketSize());
		len = read(fd, packet->Data(), packet->Size());
		if (len <= 0) {
			delete packet;
			return (NULL);
		}
		packet->SetSize(len);
	} catch(bad_alloc) {
		if (!nonblock) {
			/*
			 * If we are blocking, must do a blocking read
			 */
			if (toss(fd) <= 0) {
				return (NULL);
			}
		}
		if (packet != NULL) {
			/*
			 * A packet was allocated, but SetSize failed
			 */
			delete packet;
		}
		return (NULL);
	}
	return (packet);
}

/*
 * Here we attempt to read as many packets as we can without
 * blocking from the driver
 */
void
AtmDevice::fill_packetlist(void)
{
	BNetPacket *packet;

	/*
	 * Turn on nonblocking
	 */
	nonblock = 1;
	ioctl(fd, STUSB_NONBLOCK, &nonblock, sizeof(nonblock));


	/*
	 * And keep reading packets until we would block (but don't
	 * actually block). 
	 */
	for (;;) {
		packet = read_packet();
		if (packet == NULL) {
			return; /* no more packets or out of memory */
		}

		try {
			packetlist.ChainToEnd(packet);
		} catch (bad_alloc) {
			delete packet;
			return; /* out of memory: stop reading */
		}
	}
}

BNetPacket *
AtmDevice::ReceivePacket(void)
{
	BNetPacket *packet;

	packet = packetlist.RemoveHead();
	if (packet) {
		return (packet);
	}
	if (nonblock) {
		nonblock = 0;
		ioctl(fd, STUSB_NONBLOCK, &nonblock, sizeof(nonblock));
	}

	/*
	 * Wait for a packet
	 */
	do {
		packet = read_packet(); /* NULL if out of memory */
	} while (packet == NULL); 


	/*
	 * Got a packet, now check for more
	 */
	fill_packetlist();

	return (packet);
}


