// BeOS USB driver for Alcatel Speed Touch USB ADSL modem
// Authors: Philippe Houdoin <philippehoudoin@free.fr>
//          Maurice Theobald <maurice@theobald.com>
//
// This driver is based on ALCATEL SpeedTouch USB modem microcode
// upload & ADSL link UP utility source code for Linux & BSD,
// Copyright (C) 2001 Benoit PAPILLAULT <benoit.papillault@free.fr>, 
// Francois ROGLER <francois@rogler.org>, Richard TOBIN <richard@cogsci.ed.ac.uk>
//  
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

// DONE
// ----
// - USB detection (configuration/interfaces checked)
// - MAC Address extraction
// - Firmware upload at init
// - Get modem status
// - Start line synchronization
//
// TODO
// ----
// - Read
// - Write

#include <stdio.h>

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ByteOrder.h>
#include <USB.h>

#define DEBUG 1
#include <Debug.h>

#include "driver.h"
#include "sar.h"
#include "speed_touch_ctl.h"
#include "USB_langIDs.h"		// USB language IDs
#include "speed_touch_usb.h"	// Alcatel Speed Touch USB related definitions

#ifdef CODEWARRIOR
	#pragma mark [Local definitions]
#endif

// Local definitions
// -----------------

#ifndef DRIVER_NAME
#define DRIVER_NAME 		    "generic"
#endif
#define DEBUGGER_COMMAND	 	DRIVER_NAME


// globals for driver instances
static usb_module_info        *g_usb = NULL;
static char *				g_driver_name = DRIVER_NAME;

// devices list of devices instances (single linked list)
static sem_id				g_device_list_lock = -1;		// devices list locker 
static device *	 			g_device_list = NULL;			// head pointeur, see 'next' in device type 
static char **				g_device_name_list = NULL;		// dynamically compute in publish_devices()


// Prototypes of device hooks functions
static status_t device_added(const usb_device * usb_dev, void ** cookie);
static status_t device_removed(void * cookie);

static status_t device_open(const char *name, uint32 flags, void ** cookie);
static status_t device_close(void * cookie);
static status_t device_free(void * cookie);
static status_t device_control(void * cookie, uint32 msg,void * buf, size_t len);
static status_t device_read(void * cookie, off_t pos, void * buf, size_t * len);
static status_t device_write(void * cookie, off_t pos, const void * buf, size_t * len);



/* And, so, the device hooks list */
static device_hooks g_device_hooks = {
	device_open, 			/* -> open entry point */
	device_close,         	/* -> close entry point */
	device_free,          	/* -> free entry point */
	device_control,       	/* -> control entry point */
	device_read,       	  	/* -> read entry point */
	device_write,        	/* -> write entry point */
	NULL,               	/* -> select entry point */
	NULL,               	/* -> deselect entry point */
	NULL,               	/* -> readv */
	NULL                	/* -> writev */
};

// For USB module
static usb_notify_hooks g_usb_notify_hooks = {
	device_added,
	device_removed
};

static usb_support_descriptor g_usb_supported_devices[] = {
	{ 0, 0, 0, ALCATEL_VENDOR_ID, STUSB_PRODUCT_ID}
};

#ifdef CODEWARRIOR
	#pragma mark [Local prototypes]
#endif

// Privates prototypes
// -------------------

static long			device_count();
static device * 	device_find_by_name(const char * name);
static status_t 	device_init(device * dev);
static status_t		device_uninit(device * dev);

static void 		device_status_callback(void * cookie, uint32 usb_status, void * data, uint32 actual_len);
static void 		device_read_callback(void * cookie, uint32	usb_status,	void * data, uint32 actual_len);
static void 		device_write_callback(void * cookie, uint32	usb_status,	void * data, uint32 actual_len);

static status_t			upload_firmware(device * dev);
static unsigned char * 	load_firmware_file( int * size);
static unsigned char * 	extract_microcode(int *size, int *offset);
static int 				search_match(unsigned char * buf1, int len1, unsigned char * buf2, int len2);

static status_t		get_reference(device * dev);
static status_t		configure_modem(device * dev);
static status_t		start_line_sync(device * dev);
static status_t 	wait_line_sync(device * dev, bigtime_t timeout);
static status_t		get_modem_status(device *dev, uint8 * buf);
static status_t		get_modem_statistics(device *dev, uint8 * buf);
static void			print_line_state(uint8 * buf);
static int32        stusbPollingThread(void *cookie);

void 		dump_data(uint8 * data, int len);


#ifdef DEBUGGER_COMMAND
	static int 		device_debugger_command(int argc, char ** argv);
#endif

static status_t 	device_clear_statistics(device * dev);

#define ROUND_UP(x, y) (((x) + (y) - 1) & ~((y) - 1))


// OKAY, NOW IMPLEMENTATION PLEASE!
// --------------------------------

#ifdef CODEWARRIOR
	#pragma mark [Driver API calls]
#endif

// Driver API calls
// ----------------


_EXPORT int32 api_version = B_CUR_DRIVER_API_VERSION;


// -------------------------------------------------------
_EXPORT status_t init_hardware (void)
{
	DEBUG_ONLY( dprintf(LOGID "--- init_hardware\n") );
	return B_OK;
}

//TEMP DEBUG
uint8 dbg_buf[32];

// -------------------------------------------------------
_EXPORT status_t init_driver(void)
{
	status_t	status;

	DEBUG_ONLY( dprintf(LOGID "--- init_driver, built %s %s\n", __DATE__, __TIME__) );

#ifdef DEBUG
	if ( load_driver_symbols(DRIVER_NAME) != B_OK )
		dprintf(LOGID WARN "load_driver_symbols(\"" DRIVER_NAME "\") FAILED !\n");
	else
		dprintf(LOGID "load_driver_symbols(\"" DRIVER_NAME "\") done.\n");
#endif

	status = get_module(B_USB_MODULE_NAME, (module_info **) &g_usb);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_module(\"" B_USB_MODULE_NAME "\") FAILED: %s\n",
							strerror(status)) );
		return status;
		};

	if ((g_device_list_lock = create_sem (1, DRIVER_NAME " device_list_lock")) < 0)
		{
		put_module (B_USB_MODULE_NAME);
		return g_device_list_lock;		// error code
		};

	set_sem_owner(g_device_list_lock, B_SYSTEM_TEAM);
		
	g_usb->register_driver(g_driver_name, g_usb_supported_devices, 1, NULL);
	g_usb->install_notify(g_driver_name, &g_usb_notify_hooks);
	
#ifdef DEBUGGER_COMMAND
	add_debugger_command(DEBUGGER_COMMAND, device_debugger_command, DRIVER_NAME " driver info");
#endif	

	return B_OK;
}


// -------------------------------------------------------
_EXPORT void uninit_driver(void)
{
	long 		i;
	device *	dev;
	device * 	next_dev;

	DEBUG_ONLY( dprintf(LOGID "--- uninit_driver\n") );

	if ( g_usb != NULL )
		g_usb->uninstall_notify(g_driver_name);
	g_usb = NULL;
	
	put_module(B_USB_MODULE_NAME);

#ifdef DEBUGGER_COMMAND
	remove_debugger_command(DEBUGGER_COMMAND, device_debugger_command);
#endif

	// Free device list
	dev = g_device_list;
	while ( dev )
		{
		next_dev = dev->next;
		
		device_uninit(dev);
		free(dev);
		
		dev = next_dev;
		};

	// Free device name list
	if( g_device_name_list )
		{
		for( i = 0; g_device_name_list[i]; i++ )
			free((char *) g_device_name_list[i]);
		free(g_device_name_list);
		};
	
	delete_sem(g_device_list_lock);
}


// -------------------------------------------------------
_EXPORT const char ** publish_devices(void)
{
	long 	i;

	DEBUG_ONLY( dprintf(LOGID "--- publish_devices\n") );

	// free previous device names list
	if( g_device_name_list )
		{
		for( i = 0; g_device_name_list[i]; i++ )
			free((char *) g_device_name_list[i]);
		free(g_device_name_list);
		};
		
	// build device names list
	acquire_sem(g_device_list_lock);	
	g_device_name_list = (char **) malloc(sizeof(char*) * (device_count() + 1));	
	if( g_device_name_list )
		{
		device * dev;
		
		// iterate each device in list, and build his name...
		i = 0;
		dev = g_device_list;
		while ( dev )
			{
			g_device_name_list[i] = (char *) malloc(strlen(dev->name) + 1);
			if( g_device_name_list[i] )
				{
				strcpy(g_device_name_list[i], dev->name);
				DEBUG_ONLY( dprintf(LOGID "%s\n", g_device_name_list[i]) );
				i++;
				};
			
			// next device
			dev = dev->next;
			};

		// end the device names list
		g_device_name_list[i] = NULL;
		};

	release_sem(g_device_list_lock);
	
	return (const char **) g_device_name_list;
}


// -------------------------------------------------------
_EXPORT device_hooks * find_device
	(
	const char *	name
	)
{
	DEBUG_ONLY( dprintf(LOGID "--- find_device %s\n", name) );
	return &g_device_hooks;
}


#ifdef CODEWARRIOR
	#pragma mark [Device hooks]
#endif


// Device hooks
// ------------

// -------------------------------------------------------
static status_t device_added
	(
	const usb_device *	ud,
	void **				cookie
	)
{
	const usb_configuration_info * 	uci;
	const usb_interface_info *		uii;
	const usb_interface_info *		firmware_interface;
	const usb_endpoint_info *		firmware_in_endpoint;
	const usb_endpoint_info *		firmware_out_endpoint;
	const usb_interface_info *		data_interface;
	const usb_endpoint_info *		data_in_endpoint;
	const usb_endpoint_info *		data_out_endpoint;
	const usb_interface_info *		status_interface;
	const usb_endpoint_info *		status_int_endpoint;
	device *						dev;
	status_t						status;
	int								c;
	int								i;
	int								alt;
	char							resource_name[B_OS_NAME_LENGTH + 1];
	size_t							sz;
	void *							buffer;
	usb_device_descriptor *			udd;
	usb_string_descriptor *			usd;



	DEBUG_ONLY( dprintf(LOGID "--- device_added\n") );

	dev = NULL;

	if ( NULL == g_usb )
		return B_ERROR;

	status = B_ERROR;

	sz = 512;
	buffer = malloc(sz);
		
	// look up for Speed Touch USB (or what seem to be one...) configuration
	c = 0;
	while ( true )	// I *know* ! A infinite loop !!! Here !!! In a driver code !!!
		{			// But, relax, there's break condition(s).
		uci = g_usb->get_nth_configuration(ud, c);
		c++;		// Be prepared for next configuration (for *continue* statement)
		if ( NULL == uci )
			break;			// First *break* condition: no more configuration

		firmware_interface 		= NULL;
		data_interface 			= NULL;
		status_interface		= NULL;
			
		for ( i = 0; i < uci->interface_count; i++)
			{
			for ( alt = 0; alt < uci->interface[i].alt_count; alt++ )
				{
				uii = &uci->interface[i].alt[alt];

				// DEBUG_ONLY( dprintf(LOGID "[interface %d, alt %d]\n", i+1, alt + 1) );

				// ---- status interface ?
				if ( uii->descr->interface_number == STUSB_STATUS_INTERFACE &&
					 uii->descr->num_endpoints == 1 )
					{
					const usb_endpoint_info * 	uei;
					int 						j;
				
					status_int_endpoint 	= NULL;

					for( j = 0; j <  uii->endpoint_count; j++ )
						{
						uei = &uii->endpoint[j];
						
						if ( (uei->descr->endpoint_address & 0x81) == 0x81 && 	// input endpoint?
							 (uei->descr->attributes & 3) == 3 )				// interrupt endpoint?
							{
							// interrupt endpoint
							// DEBUG_ONLY( dprintf(LOGID "status interrupt endpoint found at %d\n", j+1) );
							status_int_endpoint = &uii->endpoint[j];
							};
						};	// for each endpoint
						
					if ( status_int_endpoint != NULL )
						{
						// status interface found
						status_interface = uii;
						// DEBUG_ONLY( dprintf(LOGID "status interface found on conf %d, interface %d, alt %d\n", c, i+1, alt+1) );
						};
						
					};	// if status interface

				// ---- data interface ?
				if ( (uii->descr->interface_number == STUSB_DATA_INTERFACE) &&  (alt == STUSB_DATA_ALT_INTERFACE) )
					{
					const usb_endpoint_info * 	uei;
					int 						j;
				
					data_in_endpoint		= NULL;
					data_out_endpoint		= NULL;

					for( j = 0; j <  uii->endpoint_count; j++ )
						{
						uei = &uii->endpoint[j];

					
						if ( (uei->descr->endpoint_address & 0x7F) != STUSB_DATA_ENDPOINTS )
							// not the endpoints we want...
							continue;
	
						// Okay, bulk endpoint. input or output ?
						if ( (uei->descr->endpoint_address & 0x80) == 0x80 )
							{
							// input endpoint
							// DEBUG_ONLY( dprintf(LOGID "data bulk in endpoint found at %d\n", j+1) );
							data_in_endpoint = &uii->endpoint[j];
							}
						else
							{
							// output endpoint
							// DEBUG_ONLY( dprintf(LOGID "data bulk out endpoint found at %d\n", j+1) );
							data_out_endpoint = &uii->endpoint[j];
							};
						};	// for each endpoint

					if ( data_in_endpoint != NULL &&
						 data_out_endpoint != NULL )
						{
						// data interface found
						data_interface = uii;
						DEBUG_ONLY( dprintf(LOGID "data interface found on conf %d, interface %d, alt %d\n", c,data_interface->descr->interface_number, data_interface->descr->alternate_setting) );
						DEBUG_ONLY( dprintf(LOGID "data in endpoint address :0x%x\n", data_in_endpoint->descr->endpoint_address) );
						DEBUG_ONLY( dprintf(LOGID "data out endpoint address :0x%x\n", data_out_endpoint->descr->endpoint_address) );
						};
					};	// if data interface
					
				// ---- firmware interface ?
				if ( uii->descr->interface_number == STUSB_FIRMWARE_INTERFACE &&
					 uii->descr->num_endpoints == 2 )
					{
					const usb_endpoint_info * 	uei;
					int 						j;
				
					firmware_in_endpoint 	= NULL;
					firmware_out_endpoint 	= NULL;

					for( j = 0; j <  uii->endpoint_count; j++ )
						{
						uei = &uii->endpoint[j];
						
						// Okay, bulk endpoint. input or output ?
						if ( (uei->descr->endpoint_address & 0x80) == 0x80 )
							{
							// input endpoint
							// DEBUG_ONLY( dprintf(LOGID "firmware bulk in endpoint found at %d\n", j+1) );
							firmware_in_endpoint = &uii->endpoint[j];
							}
						else
							{
							// output endpoint
							// DEBUG_ONLY( dprintf(LOGID "firmware bulk out endpoint found at %d\n", j+1) );
							firmware_out_endpoint = &uii->endpoint[j];
							};
						};	// for each endpoint
						
					if ( firmware_in_endpoint != NULL &&
						 firmware_out_endpoint != NULL )
						{
						// firmware interface found
						firmware_interface = uii;
						// DEBUG_ONLY( dprintf(LOGID "firmware interface found on conf %d, interface %d, alt %d\n", c, i+1, alt+1) );
						};
						
					};	// if firmware interface

				}; // next alternate interface settings
			}; // next interface

		// Check this configuration for ACM?
		
		if ( firmware_interface == NULL ||		// firmware interface not found ?
			 firmware_in_endpoint == NULL || 	// firmware bulk in endpoint not found ?
			 firmware_out_endpoint == NULL || 	// firmware bulk out endpoint not found ?
			 data_interface == NULL ||			// data interface not found ?
			 data_in_endpoint == NULL ||		// data bulk in endpoint not found ?
			 data_out_endpoint == NULL ||		// data bulk out endpoint not found ?
			 status_interface == NULL ||		// status interface not found ?
			 status_int_endpoint == NULL )		// status interrupt endpoint not found ?
			// Not a (fully-)STUSB configuration found, maybe next one?
			continue;

			
		status = g_usb->set_alt_interface(ud, data_interface);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(LOGID ERR "set_alt_interface(data_interface/%d) FAILED (%d)!\n",
						data_interface->descr->interface_number, status) );
			continue;	// Maybe next configuration is the one?
			};

		status = g_usb->set_configuration(ud, uci);
		if ( status != B_OK )
			{
			// Can't set this configuration !!!
			DEBUG_ONLY( dprintf(LOGID ERR "set_configuration(%d) FAILED (%d)!\n", c, status) );
			continue;	// Maybe next (if any) configuration is the one?
			};

		DEBUG_ONLY( dprintf(LOGID "ADSL ALCATEL SpeedTouch USB modem found!\n") );

		dev = (device *) malloc(sizeof(*dev) + 4);
		if ( dev == NULL )
			{
			// Can't allocate memory for this device!!!
			// We abort, because even if another configuration can be a good one,
			// we hardly would success to allocate again, so why persist??? 
			DEBUG_ONLY( dprintf(LOGID ERR "Can't allocate(device). ABORT!\n") );
			status = B_NO_MEMORY;
			goto failed;
			};
	
		// Init this device block/cookie with invalid values, for failure exit
		memset(dev, 0, sizeof(*dev));

		dev->open_count = 0;
		
		dev->read_semaphore 	= -1;
		dev->write_semaphore 	= -1;
		dev->pdu_read_semaphore = -1;

		dev->buffer_area_id 	= -1;
	
		// default VPI/VCI
		dev->vpi                = DEFAULT_VPI;
		dev->vci                = DEFAULT_VCI;
		
		dev->firmware_interface 	= firmware_interface;
		dev->firmware_in_endpoint 	= firmware_in_endpoint;
		dev->firmware_out_endpoint 	= firmware_out_endpoint;

		dev->data_interface 		= data_interface;
		dev->data_in_endpoint 		= data_in_endpoint;
		dev->data_out_endpoint 		= data_out_endpoint;
		
		dev->status_interface		= status_interface;
		dev->status_int_endpoint	= status_int_endpoint;

		// Okay, setup this device block/cookie		
		dev->ud = ud;

	{
	uint8					manufacturer = 0;
	uint8					product = 0;
	uint8					serial_number = 0;

	memset(buffer, 0, sz);
		
	status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_DEVICE, 0, 0, buffer, sz, &sz);
	if (status != B_OK)
		dprintf(LOGID ERR "get_descriptor(USB_DESCRIPTOR_DEVICE) FAILED (%d)!\n", status);
	else
		{
		udd = buffer;
		manufacturer 	= udd->manufacturer;
		product			= udd->product;
		serial_number	= udd->serial_number;
		};

#if 0
	sz = 512;
	memset(buffer, 0, sz);

	#define USB_LANGIDS	(0)

	status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING, USB_LANGIDS, 0, buffer, sz, &sz);
	if (status != B_OK)
		dprintf(LOGID ERR "get_descriptor(USB_LANGIDS) FAILED (%d)!\n", status);
	else
		{
		uint16 *	langid;
		int			i;
	
		usd = buffer;
		langid = (uint16 *) usd->string;
		
		dprintf(LOGID " LANGIDS supported:\n");
		for( i = 0; i < ((usd->length - 2) / 2); i++)
			{
			dprintf(LOGID " 0x%x\n", *langid);
			langid++;
			};
		};

	if ( manufacturer )
		{
		sz = 512;
		memset(buffer, 0, sz);

		status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING,
						manufacturer, USB_LANGID_ENGLISH_UNITED_STATES, buffer, sz, &sz);
		if (status != B_OK)
			dprintf(LOGID ERR "get_descriptor(manufacturer) FAILED (%d)!\n", status);
		else
			{
			char 	text[256];
			int i;
			
			usd = buffer;
			for (i=0; i < (usd->length/2); i++)
				text[i] = usd->string[2*i];	// UNICODE string
			text[i] = 0;
			dprintf(LOGID " manufacturer: [%s]\n", text);
			};
		};

	if ( product )
		{
		sz = 512;
		memset(buffer, 0, sz);

		status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING,
						product, USB_LANGID_ENGLISH_UNITED_STATES, buffer, sz, &sz);
		if (status != B_OK)
			dprintf(LOGID ERR "get_descriptor(product) FAILED (%d)!\n", status);
		else
			{
			char 	text[256];
			int i;
			
			usd = buffer;
			for (i=0; i < (usd->length/2); i++)
				text[i] = usd->string[2*i];	// UNICODE string
			text[i] = 0;
			dprintf(LOGID " product: [%s]\n", text);
			};
		};
#endif

	if ( serial_number )
		{
		sz = 512;
		memset(buffer, 0, sz);

		status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING,
						serial_number, USB_LANGID_ENGLISH_UNITED_STATES, buffer, sz, &sz);
		if (status != B_OK)
			dprintf(LOGID ERR "get_descriptor(manufacturer) FAILED (%d)!\n", status);
		else
			{
			char 	text[256];
			int i;
			
			usd = buffer;
			for (i=0; i < (usd->length/2); i++)
				text[i] = usd->string[2*i];	// UNICODE string
			text[i] = 0;
			// dprintf(LOGID " serial_number: [%s]\n", text);
			
#define HEX2INT(c) ( (c >= '0')&&(c <= '9') ?  (c - '0') : ((c & 0xf)+9) )

			for (i=0; i<6; i++)
			    dev->mac_address.byte[i] = (HEX2INT(text[i*2])*16) + (HEX2INT(text[i*2+1]));

			dprintf(LOGID "MAC address: %02x:%02x:%02x:%02x:%02x:%02x\n",
					dev->mac_address.byte[0], dev->mac_address.byte[1],
					dev->mac_address.byte[2], dev->mac_address.byte[3],
					dev->mac_address.byte[4], dev->mac_address.byte[5] );

			};
		};

	}

		// Compute device unique id
		acquire_sem(g_device_list_lock);
		dev->id = device_count();
		release_sem(g_device_list_lock);
		sprintf(dev->name, DEVICE_PATH, dev->id);

		// -- Okay, okay okay, well done!

		// Add this new device to device (linked) list 					
		acquire_sem(g_device_list_lock);
		dev->next = g_device_list;	// No list order, so we add new device at list head
		g_device_list = dev;
		release_sem(g_device_list_lock);

		free(buffer);

		*cookie = dev;
		return B_OK;	// Second *break* condition: device fully detected
		};	// while ( configuration )...

failed:;

	free(buffer);

	if ( dev != NULL )
		free(dev);

	// Glup, this device isn't a (at least supported) Speed Touch USB device
	return status;
}


// -------------------------------------------------------
static status_t device_removed
	(
	void *			cookie
	)
{
	device *	dev;
	
	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_removed: COOKIE BURNED !\n") );
		return EINVAL;
		};


	// remove it from the list of devices	
	acquire_sem(g_device_list_lock);
	if(dev == g_device_list)
		{
		g_device_list = dev->next;
		}
	else
		{
		device * tmp;
		tmp = g_device_list;
		while ( tmp )
			{
			if ( tmp->next == dev )
				{
				// remove this device from the device (linked) list
				tmp->next = dev->next;
				break;
				};
			tmp = tmp->next;
			};
		};
	release_sem(g_device_list_lock);
	
	// Okay, now try to free device
	if ( atomic_add(&(dev->open_count), -1) == 0 )
		{
		// not (still) open, we can free this device now
		device_uninit(dev);
		DEBUG_ONLY( dprintf(LOGID "--- device_removed %s\n", dev->name) );
		free(dev);
		}
	else
		// still open, device_free hook will free this device
		DEBUG_ONLY( dprintf(LOGID "--- device_removed: %s still open, marked for removal\n", dev->name) );
	
	return B_OK;
}


// -------------------------------------------------------
static status_t device_open
	(
	const char *	name, 
	uint32 			flags, 
	void **			cookie
	)
{
	device *	dev;
	status_t	status;

	dev = device_find_by_name(name);
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_open: %s NOT FOUND, ARGH !\n", name) );
		return EINVAL;
		};
		
	if ( atomic_add(&(dev->open_count), 1) == 0 )	// atomic_xxx() return previous value
		{
		// First open on this device :
		status = device_init(dev);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(LOGID ERR "--- device_open: device_init(%s)->%d, FAILURE !\n", dev->name, status) );
			return status;
			};
		}
	else
		{
			// device already open ! sorry
			DEBUG_ONLY( dprintf(LOGID "--- device_open: %s already opened, sorry.\n", dev->name) );
			return B_BUSY;
		};

	dev->flags = flags;
#if 1
	DEBUG_ONLY( dprintf(LOGID "--- device_open %s (%s%s)\n",
						dev->name,
						( ((dev->flags & O_RWMASK) == O_RDONLY) ? "O_RDONLY" :
						  ((dev->flags & O_RWMASK) == O_WRONLY) ? "O_WRONLY" : "O_RDWR"),
						(dev->flags & O_NONBLOCK) ? " O_NONBLOCK" : "") );
#endif						
	*cookie = dev;

	return B_OK;
}


// -------------------------------------------------------
static status_t device_close
	(
	void *	cookie
	)
{
	device *	dev;
	status_t	status;
	status_t    rc;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_close: COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(LOGID "--- device_close %s\n", dev->name) );
	atomic_add(&(dev->open_count), -1);

	if((dev->open_count == 0))
	{
		dev->pollingOn  = false;
		wait_for_thread( dev->pollingTid, &rc );
		stop_SAR_service(dev);
	}

	return B_OK;
}


// -------------------------------------------------------
static status_t device_free
	(
	void *	cookie
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_free: COOKIE BURNED !\n") );
		return B_ERROR;
		};

	DEBUG_ONLY( dprintf(LOGID "--- device_free %s\n", dev->name) );
	
	return B_OK;
}



// -------------------------------------------------------
static status_t device_read
	(
	void *		cookie, 
	off_t 		pos, 
	void *		data, 
	size_t *	len
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_read: COOKIE BURNED !\n") );
		return B_ERROR;
		};

	if ( (dev->flags & O_RWMASK) != O_RDONLY && (dev->flags & O_RWMASK) != O_RDWR )
		// Device not open with READ permission!!!
		return B_PERMISSION_DENIED;


	status = sar_read(dev, data, len);
	return status;
}


// -------------------------------------------------------
static status_t device_write
	(
	void *			cookie, 
	off_t 			pos, 
	const void *	data, 
	size_t *		len
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_write: COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	if ( (dev->flags & O_RWMASK) != O_WRONLY && (dev->flags & O_RWMASK) != O_RDWR )
		// Device not open with WRITE permission!!!
		return B_PERMISSION_DENIED;
#if 0
	DEBUG_ONLY( dprintf(LOGID "--- device_write %s: %x, %d\n",
				dev->name, data, *len ) );
#endif

	status = sar_write(dev, data, len);
	return status;
}


// -------------------------------------------------------
static status_t device_control
	(
	void *	cookie, 
	uint32 	op_code, 
	void *	data, 
	size_t 	len
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_control: COOKIE BURNED !\n") );
		return B_ERROR;
		};

	switch ( op_code ) {
	case STUSB_NONBLOCK:
		{
		// Set to blocking i/o on/off
		if ( *((bool *) data) )
			atomic_or(&(dev->flags), (uint32) O_NONBLOCK);
		else
			atomic_and(&(dev->flags), (uint32) ~O_NONBLOCK);
		return B_NO_ERROR;
		};

		
	case STUSB_GET_MAC_ADDR:
		{
		uint8 	i;

		DEBUG_ONLY( dprintf(LOGID "--- control_hook %s: STUSB_GET_MAC_ADDR len=%d\n", dev->name, len) );
		for ( i = 0; i < 6; i++ )
			((uint8 *) data)[i] = dev->mac_address.byte[i];
		return B_NO_ERROR;
		};

	case STUSB_SET_VCC:
		{
			T_atmVcc *vcc;

			if(data == NULL)
				return B_ERROR;

			vcc = (T_atmVcc  *)data;

			dev->vpi = vcc->vpi;
			dev->vci = vcc->vci;
			DEBUG_ONLY( dprintf(LOGID "--- control_hook set vpi/vci: %d/%d\n", dev->vpi,dev->vci) );
		};
		break;
		
	case STUSB_GET_LINK_STATE:
	{
		*((uint32 *) data) = dev->up;
	}
	break;

	case STUSB_GET_FRAME_SIZE:
	{
		*((uint32 *) data) = STUSB_RXBUF_SIZE;
	}
	break;

	case STUSB_GET_STATISTICS:
	{
		// TODO
	}
	break;

	default:
		{
		DEBUG_ONLY( dprintf(LOGID WARN "--- device_control %s: op_code %d (0x%x) unhandled, , data = 0x%x, len=%d.\n",
							dev->name, op_code, op_code, data, len) );
		return EINVAL;
		};
	}
	return B_OK;
}

#ifdef CODEWARRIOR
	#pragma mark [USB tools]
#endif

// -------------------------------------------------------
status_t	usb_read
	(
	device *	dev,
	void *		data,
	size_t		sz,	
	usb_pipe *	handle,	// if NULL, use default read endpoint
	bigtime_t	timeout
	)
{
	status_t		status;

	// BEWARE : because asynchronous requests are handled in another thread, data must point to heap, not stack!

	if ( NULL == dev )
		return B_BAD_VALUE;

	if ( handle == NULL )
		handle = dev->data_in_endpoint->handle;

	dev->read_actual_length = 0;
	dev->read_status = B_USB_STATUS_SUCCESS;
		
	status = g_usb->queue_bulk(handle, data, sz, device_read_callback, dev);
	if ( status != B_OK )
		{
		dev->nb_read_errors++;
		DEBUG_ONLY( dprintf(LOGID ERR "queue_bulk(read) FAILED(%d)!\n", status) );
		return status;
		};	

	status = acquire_sem_etc(dev->read_semaphore, 1, B_RELATIVE_TIMEOUT, timeout);

	if ( status != B_OK )
		{
		// timeout or abort
		dev->nb_read_errors++;

		DEBUG_ONLY( dprintf(LOGID ERR "queue_bulk(read) %s (0x%x)!\n",
										((status == B_TIMED_OUT) ? "TIMEOUT" : "INTERRUPTED"), status ) );
		return status;
		};
		
	if ( dev->read_status != B_USB_STATUS_SUCCESS )
		{
			dev->nb_read_errors++;
			DEBUG_ONLY( dprintf(LOGID ERR "usb_read() FAILED %04x!\n", dev->read_status) );
			return status;
		};

	return B_OK;
}
	

// -------------------------------------------------------
status_t	usb_write
	(
	device *	dev,
	void *		data,
	size_t		sz,	
	usb_pipe *	handle,	// if NULL, use default write endpoint
	bigtime_t	timeout
	)
{
	status_t		status;

	// BEWARE : because asynchronous requests are handled in another thread, data must point to heap, not stack!

	if ( NULL == dev )
		return B_BAD_VALUE;

	if ( handle == NULL )
		handle = dev->data_out_endpoint->handle;

	dev->write_actual_length = 0;
	dev->write_status = B_USB_STATUS_SUCCESS;
		
	status = g_usb->queue_bulk(handle, data,
									sz, device_write_callback, dev);
	if ( status != B_OK )
		{
		dev->nb_write_errors++;
		DEBUG_ONLY( dprintf(LOGID ERR "queue_bulk(write) FAILED(%d)!\n", status) );
		return status;
		};	

	status = acquire_sem_etc(dev->write_semaphore, 1, B_RELATIVE_TIMEOUT, timeout);
	if ( status != B_OK )
		{
		// timeout or abort
		dev->nb_write_errors++;

		DEBUG_ONLY( dprintf(LOGID ERR "queue_bulk(write) %s (0x%x)!\n",
										((status == B_TIMED_OUT) ? "TIMEOUT" : "INTERRUPTED"), status ) );
		return status;
		};
		
	if ( dev->write_status != B_USB_STATUS_SUCCESS )
		{
		dev->nb_write_errors++;
		DEBUG_ONLY( dprintf(LOGID ERR "usb_write() FAILED%s!\n", dev->write_status) );
		return status;
		};

	return B_OK;
}


#ifdef CODEWARRIOR
	#pragma mark [USB callbacks]
#endif

// USB callbacks
// -------------

// -------------------------------------------------------
static void device_status_callback
	(
	void *	cookie, 
	uint32 	usb_status, 
	void *	data, 
	uint32	actual_len
	)
{
	status_t 	status;
	device *	dev;
	uint8 *		cdata;
	
	cdata = data;
		
	dev = (device *) cookie;
	if ( NULL == dev )
		{
			DEBUG_ONLY( dprintf(LOGID ERR "--- device_status_callback: COOKIE BURNED !\n") );
			return;
		};
	
	DEBUG_ONLY(	dprintf(LOGID "--- device_status_callback %s: 0x%x, 0x%x, %d\n",
	 			dev->name, usb_status, data, actual_len) );
	
	if ( usb_status == B_USB_STATUS_IRP_CANCELLED_BY_REQUEST )
		{
			DEBUG_ONLY( dprintf(LOGID WARN "--- device_status_callback: cancelled !\n") );
			// cancelled by software
			return;
		}
		
	if (usb_status != B_USB_STATUS_SUCCESS)
		{
		DEBUG_ONLY(	dprintf(LOGID ERR "--- device_status_callback %s: 0x%x, 0x%x, %d\n",
	 			   dev->name, usb_status, data, actual_len) );
	
		status = g_usb->clear_feature(dev->status_int_endpoint->handle, USB_FEATURE_ENDPOINT_HALT);
		if (status != B_OK)
			DEBUG_ONLY( dprintf(LOGID ERR "clear_feature() FAILED (%d)!\n", status) );
		}
	else
		{
		// Got a report
		// DEBUG_ONLY( dump_data(cdata, actual_len) );
		
		// 6 bytes are received when the line goes UP : a1 00 01 00 00 00
		// 6 bytes are received for DOWN              : a1 00 00 00 00 00
		if ( actual_len >= 6 && cdata[0] == 0xA1 )
			{
				dev->up = true;
			};
		};

	// Issue next request
	status = g_usb->queue_interrupt(dev->status_int_endpoint->handle, dev->status_buffer,
									dev->status_int_endpoint->descr->max_packet_size,
									device_status_callback, dev);
	if (status != B_OK)
		{
		// XXX probably endpoint stall
		DEBUG_ONLY( dprintf(LOGID ERR "queue_interrupt() FAILED (%d)!\n", status) );
		return;
		};
}



// -------------------------------------------------------
static void device_read_callback
	(
	void *	cookie, 
	uint32 	usb_status, 
	void *	data, 
	uint32	actual_len
	)
{
	device *	dev;
		
	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_read_callback: COOKIE BURNED !\n") );
		return;
		};
#if 0	
	DEBUG_ONLY(	dprintf(LOGID "--- device_read_callback %s: status 0x%x, data 0x%x, len %d \n",
	  			dev->name, usb_status, data, actual_len) );
#endif
	dev->read_status		= usb_status;
	dev->read_actual_length = actual_len;

	if ( usb_status == B_USB_STATUS_IRP_CANCELLED_BY_REQUEST )
	{
		DEBUG_ONLY( dprintf(LOGID WARN "--- device_read_callback: cancelled !\n") );
	}
	if (usb_status == B_USB_STATUS_SUCCESS)
		{
		dev->unread_bytes 		+= actual_len;
		dev->nb_bytes_received 	+= actual_len;
#if 0
#if DEBUG
		dprintf(LOGID "--- device_read_callback %s: status 0x%x, data 0x%x, len %d \n",
	  			dev->name, usb_status, data, actual_len);

		dump_data(data, actual_len);
#endif
#endif
		};

	release_sem(dev->read_semaphore);		// Okay, notify end of bulk read (IN) request
}


// -------------------------------------------------------
static void device_write_callback
	(
	void *	cookie, 
	uint32 	usb_status, 
	void *	data, 
	uint32	actual_len
	)
{
	status_t 	status;
	device *	dev;
		
	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_write_callback: COOKIE BURNED !\n") );
		return;
		};
#if 0
	 DEBUG_ONLY(	dprintf(LOGID "--- device_write_callback %s: status 0x%x, data 0x%x, len %d \n",
	 				dev->name, usb_status, data, actual_len) );
#endif
	dev->write_status			= usb_status;
	dev->write_actual_length 	= actual_len;

	release_sem(dev->write_semaphore);	// Okay, notify end of bulk write (OUT) request
}


#ifdef CODEWARRIOR
	#pragma mark [Privates routines]
#endif

// Privates routines
// -----------------

// -------------------------------------------------------
static long device_count()
{
	device *	dev;
	long		count;
	
	count = 0;
	dev = g_device_list;
	while ( NULL != dev )
		{
		count++;
		dev = dev->next;
		};

	return count;
}

// -------------------------------------------------------
static device * device_find_by_name
	(
	const char *	name
	)
{
	device *	dev;
	
	acquire_sem(g_device_list_lock);
	dev = g_device_list;
	while ( dev )
		{
		if ( strcmp(dev->name, name) == 0 )
			{
			// name found!
			release_sem(g_device_list_lock);
			return dev;
			};

		// next one, maybe?
		dev = dev->next;
		};

	release_sem(g_device_list_lock);	
	return NULL;
}

// -------------------------------------------------------
static status_t device_init
	(
	device *	dev
	)
{
	status_t		status;

	if ( NULL == dev )
		return B_BAD_VALUE;
		
	DEBUG_ONLY( dprintf(LOGID "--- device_init %s\n", dev->name) );

	// init startup values
	dev->flags					= 0;

	dev->read_semaphore 		= create_sem(0, LOGID "read_sem");
	dev->unread_bytes			= 0;
	dev->pdu_read_semaphore     = create_sem(0, LOGID "pdu_read_semaphore");
	dev->write_semaphore 		= create_sem(0, LOGID "write_sem");
	dev->status_semaphore 		= create_sem(0, LOGID "status_sem");

	set_sem_owner(dev->read_semaphore, B_SYSTEM_TEAM);
	set_sem_owner(dev->write_semaphore, B_SYSTEM_TEAM);
	set_sem_owner(dev->status_semaphore, B_SYSTEM_TEAM);
	set_sem_owner(dev->pdu_read_semaphore, B_SYSTEM_TEAM);
	
	dev->status_buffer = malloc(64*53);
	dev->up = false;

	device_clear_statistics(dev);
	
	status = upload_firmware(dev);
	if ( status != B_OK )
		{
			dprintf(LOGID ERR "Firmware upload failed !\n");
			return( status );
		}

	// Okay, fire status interrupt endpoint now
	status = g_usb->queue_interrupt(dev->status_int_endpoint->handle, dev->status_buffer,
									dev->status_int_endpoint->descr->max_packet_size,
									device_status_callback, dev);
	if (status != B_OK)
		{
		DEBUG_ONLY( dprintf(LOGID ERR "queue_interrupt() FAILED (%d)!\n", status) );
		return status;
		};
	DEBUG_ONLY( dprintf(LOGID "Status interrupt endpoint monitor started.\n") );
	
	status = get_reference(dev);
	if ( status != B_OK )
		{
			return( status );
		}

	status = configure_modem(dev);
	if ( status != B_OK )
		{
			return( status );
		}
		
	status = start_line_sync(dev);
	if ( status != B_OK )
		{
			return( status );
		}

	status = wait_line_sync(dev, 60000000L);	// 60 seconds max
	if ( status != B_OK )
		{
			return( status );
		}

	dev->pollingTid = spawn_kernel_thread(stusbPollingThread, "stusbPollingThread", 100, dev);
	dev->pollingOn  = true;
	status = resume_thread (dev->pollingTid);
	if(status != B_OK )
	{
		return(status);
	}

	// segmentation en reassembly
	status = start_SAR_service( dev );	

	return status;
}


// -------------------------------------------------------
static status_t device_uninit
	(
	device *	dev
	)
{
	if ( NULL == dev )
		return B_BAD_VALUE;

	DEBUG_ONLY( dprintf(LOGID "--- device_uninit %s\n", dev->name) );

	if ( dev->status_buffer )
		{
		g_usb->cancel_queued_transfers(dev->status_int_endpoint->handle);
		free(dev->status_buffer);
		};

	delete_sem(dev->read_semaphore);
	delete_sem(dev->write_semaphore);
	delete_sem(dev->status_semaphore);
	delete_sem(dev->pdu_read_semaphore);

	return B_OK;
}



// -------------------------------------------------------
static status_t device_clear_statistics
	(
	device * dev
	)
{
	if ( NULL == dev )
		return B_BAD_VALUE;

	dev->nb_bytes_sent 		= 0;
	dev->nb_bytes_received 	= 0;
	dev->nb_request_failure	= 0;
	dev->nb_write_errors	= 0;
	dev->nb_read_errors		= 0;
	
	return B_OK;
};



// -------------------------------------------------------
static status_t upload_firmware
	(
	device *	dev
	)
{
	status_t	status;
	uint8 * 	data;
	size_t		sz;
	uint8 *     firmware;
	int         firmware_size;
	int         firmware_offset;
	uint8 *		ptr;
	int 		total;
	int			written;

	status 		= B_ERROR;

	data = NULL;
	firmware = extract_microcode(&firmware_size, &firmware_offset);
	if(firmware == NULL)
	  	{
	    dprintf(LOGID ERR "No firmware: ABORTING...\n");
	    goto exit;
	  	};

	DEBUG_ONLY( dprintf(LOGID "Firmware offset %d, size %d\n", firmware_offset, firmware_size) );

	// allocate generic data buffer for read and write...
	sz 			= 0x1000;
	data 		= malloc(sz + 16);	// add some padding bytes
	
	// Halt/reset firmware endpoints
	status = g_usb->clear_feature(dev->firmware_out_endpoint->handle, USB_FEATURE_ENDPOINT_HALT);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "clear_feature(firmware_out) FAILED(%d)!\n", status) );
		goto exit;
		};
		
	status = g_usb->clear_feature(dev->firmware_in_endpoint->handle, USB_FEATURE_ENDPOINT_HALT);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "clear_feature(firmware_in) FAILED(%d)!\n", status) );
		goto exit;
		};	

	DEBUG_ONLY( dprintf(LOGID "Firmware endpoints reset done.\n") );

 	// URB 8 : both leds are static green
	sz = sizeof(block_1);
	memcpy(data, block_1, sz);

	status = usb_write(dev, data, sz, dev->firmware_out_endpoint->handle, STUSB_TIMEOUT);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "URB8 write FAILED%s!\n", status) );
		goto exit;
		};

	DEBUG_ONLY( dprintf(LOGID "Firmware upload initialization...\n") );
	

	// now, USB  led : blinking green
	//      ADSL led : off
	// URB 11
	sz = 0x200;
	memset(data, 0, sz);

	status = usb_read(dev, data, sz, dev->firmware_in_endpoint->handle, STUSB_TIMEOUT);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "URB11 read FAILED%s!\n", status) );
		goto exit;
		};

	// DEBUG_ONLY( dprintf(LOGID "BLOCK2 : %d bytes downloaded : OK\n", dev->read_actual_length) );
	// DEBUG_ONLY( dump_data(data, dev->read_actual_length) );

	// Wait a little
	snooze(5000);

    // before downloading microcode :
	// + USB  led : blinking
	// + ADSL led : off
	//
	// during downloading:
	// + USB  led : blinking green
	// + ADSL led : off
	//
	// after downloading microcode :
	// + USB  led : static green
	// + ADSL led : static red

	DEBUG_ONLY( dprintf(LOGID "Uploading firmware...\n") );
		
	total = written = 0;

	ptr = firmware;
	ptr += firmware_offset;

	while ( total < firmware_size )
		{
		sz = 0x1000;
		if ( (firmware_size - total) < sz )
			sz = firmware_size - total;

		memcpy(data, ptr, sz);	

		status = usb_write(dev, data, sz, dev->firmware_out_endpoint->handle, STUSB_TIMEOUT);

		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(LOGID ERR "Firmware write FAILED at %d/%d -> 0x%x!\n",
						written + sz, firmware_size, dev->write_status) );
			goto exit;
			};
			
		ptr 	+= sz;
		total 	+= sz;
		written += dev->write_actual_length;
		};

	free(firmware);
	firmware = NULL;
		
	// DEBUG_ONLY( dprintf(LOGID "BLOCK3: %d/%d bytes uploaded : OK\n", written, firmware_size) );
 
	// URB 142

	sz = 0x200;
	memset(data, 0, sz);
	
	status = usb_read(dev, data, sz, dev->firmware_in_endpoint->handle, STUSB_TIMEOUT);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "URB142 read FAILED%s!\n", status) );
		goto exit;
		};

	// DEBUG_ONLY( dprintf(LOGID "BLOCK4 : %d bytes downloaded : OK\n", dev->read_actual_length) );
	// DEBUG_ONLY( dump_data(data, dev->read_actual_length) );

	status = B_ERROR;

	if ( written == firmware_size )
		{
		DEBUG_ONLY( dprintf(LOGID "Firmware upload done.\n") );
		status = B_OK;
		}
	else
		DEBUG_ONLY( dprintf(LOGID ERR "Firmware upload failure!\n") );

	// wait a little...
	snooze(3000000);

	// let's exit now...

exit:;

	if ( data )
		free(data);
		
	if ( firmware )
		free(firmware);
	
	return status;
}


// -------------------------------------------------------
static unsigned char * load_firmware_file( int * size)
{
	struct stat statbuf;
	int len ;
	unsigned char * buf;
	int fd;
	int i;
	char *firmware_files[] =
  		{
		"/boot/beos/etc/stusb_firmware",
    	"/boot/beos/etc/alcaudsl.sys",
		"/boot/beos/etc/mgmt.o",
		NULL
  		};

	fd = -1;
  	i  = 0;

	while ( firmware_files[i] )
		{
		// DEBUG_ONLY( dprintf(LOGID "Trying firmware %s...\n", firmware_files[i]) );
		fd = open(firmware_files[i],O_RDONLY);
		if ( fd > 0)
			break;
		i++;
		};

  if (fd < 0)
    {
	DEBUG_ONLY( dprintf(LOGID ERR "No firmware file found! ABORTING...\n") );
	return NULL;
    };
  
  if (fstat(fd,&statbuf) != 0)
    {
	DEBUG_ONLY( dprintf(LOGID ERR "Firmware file can't be fstated()!\n") );
    return NULL;
    };

   len = statbuf.st_size;
   buf = (unsigned char *) malloc(len);
   if (buf == NULL)
		{
		DEBUG_ONLY( dprintf(LOGID ERR "Can't allocate firmware buffer!\n") );
      	close (fd);
      	return NULL;
		};

   if (read(fd,buf,len) != len)
	  	{
	  	DEBUG_ONLY( dprintf(LOGID ERR "Read firmware into memory FAILED !\n") );
      	free(buf);
      	close (fd);
      	return NULL;
		};

	close(fd);

  	DEBUG_ONLY( dprintf(LOGID "Firmware file loaded: %s.\n",firmware_files[i]) );

   	*size = len;
   	return buf;
}

// -------------------------------------------------------
static unsigned char * extract_microcode(int *size, int *offset)
{
  unsigned char start_pattern[] =
  {
    0x88, 0xf8, 0x00, 0x40, 0x00, 0x00, 0xf2, 0x01, 
    0x4d, 0xea, 0x01, 0xea, 0x0d, 0xc0, 0xa0, 0xe1
  };
  
  unsigned char stop_pattern[] = 
  {
    0x01, 0x00, 0x00, 0x00, 0x40, 0x01, 0x12, 0x08,
    0x04, 0x00, 0x40, 0x00, 0x00, 0x00, 0x01, 0x14
  };
  
  int start, stop;
  unsigned char *buf1;
  int size1;
  
  if ((buf1 = load_firmware_file(&size1)) == NULL)
    return NULL;
  
  start = search_match(buf1,size1,start_pattern ,sizeof(start_pattern));
  stop  = search_match(buf1,size1,stop_pattern  ,sizeof(stop_pattern) );
  
  if (stop <= start)
    {
      free(buf1);
      return NULL;
    }
  
  *size = stop+sizeof(stop_pattern) - start;
  *offset = start;
  return (buf1);
}


// -------------------------------------------------------
static int search_match(unsigned char * buf1, int len1, unsigned char * buf2, int len2)
{
	//  search for the longest match of buf2 inside buf1. It returns an offset
	//  in buf1 for this best match. In case of errors, returns -1

	int best_match = 0; /* number of character matching */
	int best_offset = -1;

	int cur_match, cur_offset;
	int i;

	for (cur_offset=0;cur_offset<=len1-len2;cur_offset++)
	{
		// computing cur_match
		cur_match = 0;
		for (i=0;i<len2;i++)
			if (buf1[cur_offset+i] == buf2[i])
				cur_match ++;

		if (cur_match > best_match)
		{
			best_offset = cur_offset;
			best_match  = cur_match;

			if (best_match == len2)
				break;
		}
	}
	return best_offset;
}


// -------------------------------------------------------
static status_t get_reference
	(
	device *	dev
	)
{
	status_t		status;
	uint8 * 		data;
	size_t			sz;

	// allocate temporary data buffer for usb request...
	sz 			= 14;
	data 		= (uint8 *) malloc(sz + 16);	// add some padding bytes
	
	/* URB 151 : this buffer match (if you reverse one byte with its neighbour)
	   the label on the manta box!!! in my case : 3EC18607CAAB06
	3EC 18607CAAB 09 */

	status = g_usb->send_request(dev->ud,
								USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,	// device in, vendor specific request
								STUSB_GET_INFO,	STUSB_GET_REFERENCE, 0x00,
								sz, data, sz, &sz);

	if (status == B_OK)
		{
		int i, idx;
		char textbuf[64];

		textbuf[0] = '\0';
		idx = 0;
		for (i=0; i < sz; i++)
			{
			if (i==3 || i==12)
				textbuf[idx++] = ' ';
			textbuf[idx++] = data[(((i%2)==0) ? i+1 : i-1)];
			};
		textbuf[idx++] = '\0';
		
		dprintf(LOGID "Reference: %s (see under your modem box).\n", textbuf);
		}
	else
		DEBUG_ONLY( dprintf(LOGID ERR "---- get_reference: send_request() FAILED (0x%x) (len=%d)!\n", status, sz) );

	free(data);
	
	return status;
}

// -------------------------------------------------------
static status_t configure_modem
	(
	device *	dev
	)
{
	status_t		status;
	uint8 * 		data;
	size_t			sz;

	sz 			= 32;
	data 		= (uint8 *) malloc(sz + 16);	// add some padding bytes

	// 0x40 request type = USB_REQTYPE_DEVICE_OUT | USB_REQTYPE_VENDOR

	// URB 147
	data[0]=0x1c; data[1]=0x50;
	sz = 2;
	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_OUT | USB_REQTYPE_VENDOR,	// device in, vendor specific request
								0x01,	// request code (0x01 = ???)
								0x0b,	// value (0x0b = ???)
								0x00,	// index (0 = unused ?)
								sz,		// size of input data
								data,	// in/out data buffer 
								sz,		// max size of output buffer 
								&sz);	// real size of output data
	if ( status != B_USBD_SUCCESS )
		DEBUG_ONLY( dprintf(LOGID ERR "configure_modem URB147: send request status = 0x%x, len = %d\n", status, sz) );

	// URB 148
	data[0]=0x32; data[1]=0x00;
	sz = 2;
	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_OUT | USB_REQTYPE_VENDOR,	// device in, vendor specific request
								0x01,	// request code (0x01 = ???)
								0x02,	// value (0x02 = ???)
								0x00,	// index (0 = unused ?)
								sz,		// size of input data
								data,	// in/out data buffer 
								sz,		// max size of output buffer 
								&sz);	// real size of output data
	if ( status != B_USBD_SUCCESS )
		DEBUG_ONLY( dprintf(LOGID ERR "configure_modem URB148: send request status = 0x%x, len = %d\n", status, sz) );

	// URB 149
	data[0]=0x01; data[1]=0x00; data[2]=0x01;data[3]=0x00;
	sz = 3;
	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_OUT | USB_REQTYPE_VENDOR,	// device in, vendor specific request
								0x01,	// request code (0x01 = ???)
								0x03,	// value (0x03 = ???)
								0x00,	// index (0 = unused ?)
								sz,		// size of input data
								data,	// in/out data buffer 
								sz,		// max size of output buffer 
								&sz);	// real size of output data
	if ( status != B_USBD_SUCCESS )
		DEBUG_ONLY( dprintf(LOGID ERR "configure_modem URB149: send request status = 0x%x, len = %d\n", status, sz) );

	// URB 150
	data[0]=0x01; data[1]=0x00; data[2]=0x01;data[3]=0x00;
	sz = 3;
	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_OUT | USB_REQTYPE_VENDOR,	// device in, vendor specific request
								0x01,	// request code (0x01 = ???)
								0x04,	// value (0x04 = ???)
								0x00,	// index (0 = unused ?)
								sz,		// size of input data
								data,	// in/out data buffer 
								sz,		// max size of output buffer 
								&sz);	// real size of output data
	if ( status != B_USBD_SUCCESS )
		DEBUG_ONLY( dprintf(LOGID ERR "configure_modem URB150: send request status = 0x%x, len = %d\n", status, sz) );

	free(data);
	return status;
}


// -------------------------------------------------------
static status_t start_line_sync
	(
	device *	dev
	)
{
	status_t	status;
	uint8 * 	data;
	size_t		sz;

	status 		= B_ERROR;

	// allocate temporary data buffer for usb request...
	sz 			= 16;
	data 		= (uint8 *) malloc(sz);	// add some padding bytes
	
	sz = 2;
	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,	// device in, vendor specific request
								STUSB_LINE,	STUSB_START_SYNC, 0x00,
								sz, data, sz, &sz);

	if (status == B_OK)
		DEBUG_ONLY( dprintf(LOGID "Starting line sync...\n", sz) );
	else
		DEBUG_ONLY( dprintf(LOGID ERR "---- start_line_sync: send_request() FAILED (0x%x)!\n", status) );

	free(data);
	
	return status;
}



// -------------------------------------------------------
static status_t wait_line_sync
	(
	device *	dev,
	bigtime_t	timeout
	)
{
	status_t	status;
	uint8 *		buf;
	uint8 *		prevbuf;
	bigtime_t	start;
	
	buf     = malloc( 32 );
	prevbuf = malloc( 32 );	 
	
	status = get_modem_status(dev, prevbuf);
	if( status == B_OK )
		print_line_state(prevbuf);
	else
		dprintf(LOGID ERR "get_modem_status error\n");
	
	status = B_ERROR;
	start = system_time();
	for (;;)
		{
			// status polling period is 1 sec
			snooze(1000000);

			get_modem_status(dev, buf);
			if (memcmp(buf, prevbuf, TOTAL) != 0)
			{
				print_line_state(buf);
				
				if ( (buf[OFFSET_7] == 0x20 && prevbuf[OFFSET_7] != 0x20))
				{
					snooze(2000000);
					status = B_OK;
					dev->up = true;
					break;
				}
				
				if (buf[OFFSET_7] != 0x20
					&& prevbuf[OFFSET_7] == 0x20)
				{
					dprintf(LOGID "ADSL link goes DOWN\n");
					dev->up = false;
				}
				
				memcpy(prevbuf,buf,TOTAL);
			}

		if ( system_time() - start > timeout )
			{
			status = B_TIMED_OUT;
			break;
			};
		}
		
	free(buf);
	free(prevbuf);
	
	return status;
}
	

// ----------------------Modem status------------------------
static status_t get_modem_status
	(
	device *	dev,
	uint8 * 	buf
	)
{
	status_t	status;
	size_t 		sz;

	//  buf MUST BE at least 15 bytes.
	//  -1 is returned on errors, 0 otherwise

	memset(buf, 0, TOTAL);

	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,
							STUSB_LINE,	STUSB_GET_STATE_7, 0x01,
							SIZE_7, buf+OFFSET_7, SIZE_7, &sz);
	if ( status != B_USBD_SUCCESS )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_modem_status 7 = 0x%x, len = %d\n", status, sz) );
		return status;
		};
		
	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,
							STUSB_LINE,	STUSB_GET_STATE_b, 0x01,
							SIZE_b , buf+OFFSET_b , SIZE_b, &sz);
	if ( status != B_USBD_SUCCESS )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_modem_status b = 0x%x, len = %d\n", status, sz) );
		return status;
		};

	return status;
}

int32 stusbPollingThread(void *cookie)
{
	uint8 *pollbuf;
	device *dev;
	status_t status;
	
	dev = (device *)cookie;
	pollbuf = (uint8 *)malloc(16);
	while( dev->pollingOn )
	{
		snooze(1500000);
		status = get_modem_statistics(dev, pollbuf);
		if(status != B_OK)
		{
			dprintf(LOGID ERR "cannot poll modem: stusb polling thread exits\n");
			break;
		}
		dev->up = (dev->line_status == 0x20)?true:false;
	}
	free(pollbuf);
	dprintf(LOGID "stusb polling thread completed !\n");
}

// this function must be called every 1.5 seconds
status_t get_modem_statistics(device *dev, uint8 * buf)
{
	status_t	status;
	size_t 		sz;

	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,
							0x01,	STUSB_GET_STATE_e, 0x00,
							SIZE_e, buf, SIZE_e, &sz);
	if ( status != B_USBD_SUCCESS )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_modem_status e = 0x%x, len = %d\n", status, sz) );
		return status;
		};
		dev->drc += buf[0];

	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,
							0x01,	STUSB_GET_STATE_f, 0x00,
							SIZE_f, buf, SIZE_f, &sz);
	if ( status != B_USBD_SUCCESS )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_modem_status f = 0x%x, len = %d\n", status, sz) );
		return status;
		};
		dev->urc += buf[0];

	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,
							STUSB_LINE,	STUSB_GET_STATE_d, 0x01,
							SIZE_d, buf, SIZE_d, &sz);
	if ( status != B_USBD_SUCCESS )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_modem_status d = 0x%x, len = %d\n", status, sz) );
		return status;
		};
		dev->alarms = *(uint32 *)buf;

	status = g_usb->send_request(dev->ud, USB_REQTYPE_DEVICE_IN | USB_REQTYPE_VENDOR,
							STUSB_LINE,	STUSB_GET_STATE_7, 0x01,
							SIZE_7, buf, SIZE_7, &sz);
	if ( status != B_USBD_SUCCESS )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_modem_status 7 = 0x%x, len = %d\n", status, sz) );
		return status;
		};
		dev->line_status = buf[0];

	return(B_OK);
}

// -------------------------------------------------------
static void print_line_state
	(
	uint8 * buf
	)
{
	/* Displays line status*/

	uint32 down_speed, up_speed;
	
	switch (buf[OFFSET_7])
	{
    case 0x00:
		dprintf(LOGID "ADSL line is down !\n");
		break;
    case 0x08:
		dprintf(LOGID "ADSL is waiting...\n");
		break;
    case 0x10:
		dprintf(LOGID "ADSL line is initializing...\n");
		break;
    case 0x20:
    	down_speed = buf[OFFSET_b] | (buf[OFFSET_b+1]<<8)
			| (buf[OFFSET_b+2]<<16) | (buf[OFFSET_b+3]<<24);
		up_speed = buf[OFFSET_b+4] | (buf[OFFSET_b+5]<<8)
			| (buf[OFFSET_b+6]<<16) | (buf[OFFSET_b+7]<<24);
			
		dprintf(LOGID "ADSL line is up, downstream at %d kbit/s, "
			   "upstream at %d kbit/s.\n", down_speed,up_speed);
		break;
	default:
		break;
    }
}


// -------------------------------------------------------
 void dump_data
	(
	uint8 * buf,
	int 	len
	)
{
	int i,j;
  	char text[128];
	char *	ptr;

	ptr = text;
	for ( i = 0; i < len; i += 16 )
		{
      	for ( j = i; j < i+16 ; j++ )
			{
			if ( j < len )
				sprintf(ptr, "%02x ",buf[j]);
			else
				sprintf(ptr, "   ");
			ptr += 3;
			};
			
		for (j = i; j < len && j < i+16;j++)
			{
			if ( buf[j] >= ' ' && buf[j] < 0x7f )
				*ptr = buf[j];
			else
				*ptr = '.';
				
			ptr++;
			};
		*ptr = '\n';
		ptr++;
		*ptr = '\0';
		
		dprintf( text );

		// next line
		ptr = text;
		};
}


#ifdef CODEWARRIOR
	#pragma mark [Debugger Command]
#endif

#ifdef DEBUGGER_COMMAND

// -------------------------------------------------------
static int device_debugger_command
	(
	int 	argc,
	char **	argv
	)
{
	struct debugger_commands
		{
		int		id;
		char * 	name;
		char *	shortcut;
		// TO DO : command handling by sub-routine call
		// int 	(*cmd)(device * dev, int argc, char ** argv);
		char *	arguments;
		}
	* cmd, dbgr_cmds[] =
		{
			{0,  "help", "h", NULL},
			{1,  "list", "l", NULL},
			{2,  "select", "s", "{index}"},
			{3,  "stats", "st", NULL},
			{4,  "clear_stats", "cs", NULL},
			{5,  "lockers", "lck", NULL},
			{6,	 "dump_device", "dd", NULL},
			{0,  NULL, NULL, NULL}
		};
	int 			cmd_id;
	int				return_value;
	int				i;
	
	static int		dev_id = -1;
	device *		dev;

	return_value 	= 0;

	i = 0;
	dev = g_device_list;
	while ( dev )
		{
		if ( i == dev_id )
			break;
		i++;
		dev = dev->next;
		};

	if ( NULL == dev )
		kprintf("  no device selected !\n");
	else
		NULL; // w = dev->data;

	cmd_id 			= 0;	// help, by default

	if ( argc > 1 )
		{
		cmd = dbgr_cmds;
		while ( cmd->name )
			{
			if ( (strcmp(cmd->name, argv[1])== 0) ||
				 (strcmp(cmd->shortcut, argv[1]) == 0 ) )
				break;

			 cmd++;
			};

		cmd_id = cmd->id;
		};

	switch (cmd_id)
		{
		case 1: // list
			{
			device *	tmp_dev;

			i = 0;
			tmp_dev = g_device_list;
			while ( tmp_dev )
				{
				kprintf("  %d: %s", i, tmp_dev->name);
				if ( i == dev_id )
					kprintf(" (selected)");
				kprintf("\n");

				i++;
				tmp_dev = tmp_dev->next; 
				};
			break;
			};

		case 2:	// select
			{
			if ( argc >= 3 )
				{
				long	count;
				
				i = atoi(argv[2]);
				count = device_count();
				if ( (i < 0) || (i >= count) )
				 	kprintf("  index must be in [0, %d] range !\n", count - 1);
				else
					dev_id = i;
				}
			else
				kprintf("  index missing !\n");
			break;
			};
			
		case 3:	// stats
			{
			if ( NULL == dev )
				break;

		 	kprintf("  nb_bytes_sent		%d\n"
					"  nb_bytes_received	%d\n"
					"  nb_request_failure	%d\n"
					"  nb_write_errors		%d\n"
					"  nb_read_errors		%d\n",
					dev->nb_bytes_sent, dev->nb_bytes_received, dev->nb_request_failure,
					dev->nb_write_errors, dev->nb_read_errors);
			break;
			};
			
		case 4:	// clear stats
			{
			if ( NULL == dev )
				break;
			device_clear_statistics(dev);
			break;
			};

		case 5: // locker
			{
			if ( NULL == dev )
				break;

			kprintf("  device at 0x%4.4x :\n"
			        "    id                %d\n"
					"    name              %s\n"
					"    open_count        %d\n"
					"    flags             %x\n",
					dev, dev->id, dev->name, dev->open_count, dev->flags);

			DEBUG_ONLY( kprintf("    debug_level       0x%x\n", dev->debug_level) );
			break;
			};

		case 6: // dump_device
			{
			if ( NULL == dev )
				break;

			kprintf("  device at 0x%4.4x :\n"
			        "    id                %d\n"
					"    name              %s\n"
					"    open_count        %d\n"
					"    flags             %x\n",
					dev, dev->id, dev->name, dev->open_count, dev->flags);

			DEBUG_ONLY( kprintf("    debug_level       0x%x\n", dev->debug_level) );
			break;
			};
			
		case 0:	// help
		default:
			{
			kprintf(DEBUGGER_COMMAND " usage:\n");
			cmd = dbgr_cmds;
			while ( cmd->name )
				{
				kprintf("  %s", cmd->name);
				if ( cmd->shortcut )
					kprintf(" | %s", cmd->shortcut);
				if ( cmd->arguments )
					kprintf(" %s", cmd->arguments);
				kprintf("\n");

				cmd++;	// next
				};
			
			break;
			};
		};	// switch
	
	return return_value;
}

#endif // DEBUGGER_COMMAND





