// BeOS USB driver for Alcatel Speed Touch USB ADSL modem
// SAR layer

#include <stdio.h>
#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ByteOrder.h>
#include <USB.h>
#include "driver.h"
#include "sar.h"

extern void 		dump_data(uint8 * data, int len);

//--------------------------------------TYPES DECLARATION------------------------------
typedef struct rxBuf
{
	uint8        *buffer;
	struct rxBuf *next;
	size_t       pkt_len;
	bool         empty;
} rxBuf_t;

//--------------------------------------LOCAL DECLARATION------------------------------
// Rx thread identifier
static thread_id sarTid;
static int32 sarReady=0;
static rxBuf_t *rxHead;
static rxBuf_t *rxTail;
static uint8   *cellsBuffer;

//------------------------------------- FUNCTIONS PROTOTYPES --------------------------
static int32 stusbRxSarThread(void *cookie);
static status_t initMemPool();
static void freeMemPool();
static void reassemble_cells(uint8 *buffer, size_t cell_count, device *dev);

//------------------------------------- FUNCTIONS DEFININITION-------------------------
// preallocate buffers for reassembled RX PDU
status_t initMemPool()
{
	int      i;
	rxBuf_t  *ptr;
	rxBuf_t  *pfree;
	status_t status=B_OK;

	cellsBuffer    = (uint8 *)malloc( AAL5_CPCS_MAX_SIZE );
	if(cellsBuffer == NULL)
		return(B_ERROR);
	
	rxHead         = (rxBuf_t *)malloc(sizeof(rxBuf_t));
	if(rxHead == NULL)
		return(B_ERROR);
	rxHead->buffer    = (uint8 *)malloc( AAL5_CPCS_MAX_SIZE );
	if(rxHead->buffer == NULL)
	{
		free(rxHead);
		return(B_ERROR);
	}
	rxHead->empty   = true;
	rxHead->pkt_len = 0;

	ptr            = rxHead;
	for(i=0; i<RX_BUFFER_QTY-1; i++)
	{
		ptr->next   = (rxBuf_t *)malloc(sizeof(rxBuf_t));
		if(ptr->next== NULL)
		{
			status = B_ERROR;
			break;
		}

		ptr->next->buffer = (uint8 *)malloc( AAL5_CPCS_MAX_SIZE );
		if(ptr->next->buffer == NULL)
		{
			status = B_ERROR;
			break;
		}
		ptr->next->empty   = true;
		ptr->next->pkt_len = 0;
		ptr = ptr->next;	
	}
	ptr->next = NULL;
	if(status == B_OK)
	{ // all buffer are allocated
		rxTail       = ptr;
		rxTail->next = rxHead;
		rxTail       = rxHead;
		dprintf(LOGID "--- SAR mempool ready!\n");
	}
	else
	{ // not enough memory, free up already allocated
		ptr = pfree = rxHead;
		while(ptr)
		{
			if(ptr->buffer)
				free(ptr->buffer);
			pfree = ptr->next;
			free(ptr);
			ptr = pfree;
		}
		dprintf(LOGID ERR "--- unable to allocate SAR mempool!\n");
	}

	return( status );
}

// free up preallocated rx PDU buffers
void freeMemPool()
{
	rxBuf_t  *ptr;
	rxBuf_t  *pfree;
	int      i;
	
	ptr = pfree = rxHead;
	for( i=0; i<RX_BUFFER_QTY; i++)
	{
		if(ptr->buffer)
			free(ptr->buffer);
		pfree = ptr->next;
		free(ptr);
		ptr = pfree;
	}
	rxHead = rxTail = NULL;
}

status_t start_SAR_service(device *dev)
{
	status_t status;
	
	status = initMemPool();
	if(status != B_OK)
		return( status );
	sarTid = spawn_kernel_thread(stusbRxSarThread, SAR_THREAD_NAME, SAR_THREAD_PRIO, dev);
	status = resume_thread (sarTid);
	if(status != B_OK )
	{
		freeMemPool();
		return( status );
	}

	return ( status );
}

status_t stop_SAR_service(device *dev)
{
	status_t rc;
	status_t status;
	sem_info info;
	
	atomic_add(&(sarReady), -1);

	// PDU read cleanup
	get_sem_info(dev->pdu_read_semaphore, &info);
	if(info.count < 0)
	{
		release_sem(dev->pdu_read_semaphore);
	}
	
	// raw cell task read cleanup
	get_sem_info(dev->read_semaphore, &info);
	if(info.count < 0)
	{
		release_sem(dev->read_semaphore);
	}
	status = wait_for_thread( sarTid, &rc );

	// wait for end of current write transfer
	while( dev->write_on )
		snooze(STUSB_TIMEOUT);
	// free up SAR memory
	freeMemPool();
	dprintf(LOGID "---SpeedTouch USB driver SAR service stopped!\n");
	return( status );	
}

int32 stusbRxSarThread(void *cookie)
{
	uint8    *buffer;
	status_t status;
	device   *dev;
	size_t   sz;
	uint8    *cellsbuffer;
	uint8    *buf;
	uint8    *addr;

	dprintf(LOGID "---SpeedTouch USB driver SAR service started!\n");
	dev    = (device *)cookie;
	atomic_add(&(sarReady), 1);	
	buffer = (uint8 *)malloc(4096);
	
	while ( sarReady )
	{
		status = usb_read(dev, buffer, STUSB_RXBUF_SIZE, NULL, B_INFINITE_TIMEOUT);
        snooze(1000000);

		if(sz >= STUSB_RXBUF_SIZE)
		{
			dprintf(LOGID ERR "--- buffer overflow\n");
			break;
		}
		
		cellsbuffer = buffer;
		if(status != B_OK)
		{
			dprintf(LOGID ERR "--- usb_read returns 0x%x\n",status);
			continue;
		}
		sz = dev->read_actual_length;
		// if buffer size is not cell aligned drop it
		if((sz % ATM_CELL_SIZE))
		{
			int i;
			
			dprintf(LOGID ERR "--- corrupted cells buffer! (%d bytes)\n",sz);
			i           =  (sz%ATM_CELL_SIZE);
			cellsbuffer += i;
			sz          -= i;
			dump_data(buffer, i);
		}
		//else
		{
			reassemble_cells(cellsbuffer, (sz/ATM_CELL_SIZE), dev);
		}
	}
	dprintf(LOGID "---SAR thread shutdown!\n");
	free( buffer );
	exit_thread( B_OK );
}

void reassemble_cells(uint8 *buffer, size_t cell_count, device *dev)
{
	int        i;
	uint32     vpi;
	uint32     vci;
	uint32     pti;
	uint8      *cell;
	sem_info   info;
	
	if(!rxTail->empty)
	{
		dprintf(LOGID ERR "--- SAR RX ring overflow !!\n");
		dprintf(LOGID ERR "--- %d cells dropped !!\n", cell_count);
		return;
	}

	cell = buffer;	
	for(i=1; i<=cell_count; i++)
	{
		// decode cell header
		vpi = ( cell[0] << 4) | (cell[1] >> 4);
		vci = ((cell[1] & 0x0f) << 12) | (cell[2] << 4) | (cell[3] >> 4);
		pti = (cell[3] & 0x0f);
		cell += ATM_CELL_HEADER;
		// OAM cell are not processed yet
		if(vpi != dev->vpi || vci != dev->vci)
		{
			continue;
		}
		memcpy(rxTail->buffer + rxTail->pkt_len, cell, ATM_CELL_DATA);
		rxTail->pkt_len += ATM_CELL_DATA;
		if(rxTail->pkt_len >= AAL5_CPCS_MAX_SIZE)
		{
			dprintf(LOGID ERR "--- SAR RX buffer overflow !!\n");
			dprintf(LOGID ERR "--- %d cells dropped !!\n", (rxTail->pkt_len/ATM_CELL_DATA));
			rxTail->pkt_len = 0;
			return;
		}
		//check end of PDU 
		if(pti & 2)
		{
			// TODO protect critical section here ???
			rxTail->empty = false;
			rxTail        = rxTail->next;
			// END TODO
			get_sem_info(dev->pdu_read_semaphore, &info);
			if(info.count < 0)
			{
				release_sem_etc(dev->pdu_read_semaphore, 1, B_DO_NOT_RESCHEDULE);
			}
			if(!rxTail->empty)
			{
				dprintf(LOGID ERR "--- SAR RX ring overflow !!\n");
				dprintf(LOGID ERR "--- %d cells dropped !!\n", cell_count-i);
				return;
			}
		}
		// next cell
		cell  += ATM_CELL_DATA;
	}
}


// read reassembled PDU
status_t sar_read(device *dev, uint8 *buffer, uint32 *len)
{
	status_t status;
	size_t   sz;


	if(!sarReady)
		{
			dprintf(LOGID ERR "--- sar_read: SAR service unavailable !!\n");
			*len = 0;
			return (B_DEV_UNREADABLE);
		}

	if((rxHead->empty) && (dev->flags & O_NONBLOCK))
	{
		*len = 0;
		return (B_OK);
	}
	
	if(rxHead->empty)
	{
		status = acquire_sem(dev->pdu_read_semaphore);
		if(status != B_OK )
		{
			*len = 0;
			return(status);
		}
	}	

	if(!sarReady)
	{
		dprintf(LOGID ERR "--- sar_read: SAR service unavailable !!\n");
		*len = 0;
		return (B_DEV_UNREADABLE);
	}
		
	// retrieve PDU
	sz = (*len < rxHead->pkt_len)?*len:rxHead->pkt_len;
	memcpy(buffer, rxHead->buffer, sz);
	*len = sz;
	// fetch next pdu
	// TODO: protect critical section here ???
	rxHead->pkt_len = 0;
	rxHead->empty   = true;
	rxHead          = rxHead->next;
	// END TODO
	
	return(B_OK);
}


// segment PDU into cells and send it to modem
status_t sar_write(device *dev, uint8 *buffer, uint32 *len)
{
	int      i;
	int      cell_qty;
	uint8    pti;
	uint8    *cell;
	uint8    *ptr;
	uint32   sz;
	status_t status;


	if(!sarReady)
		{
			dprintf(LOGID ERR "--- sar_write: SAR service unavailable !!\n");
			*len = 0;
			return (B_DEV_UNREADABLE);
		}
					
	if((*len % ATM_CELL_DATA))
	{
		dprintf(LOGID ERR "--- bad PDU length !! (%d bytes)\n",*len);
		return(B_ERROR);
	}

	pti      = 0;
	cell_qty = *len / ATM_CELL_DATA;
	ptr  = buffer;
	cell = cellsBuffer;
	memset(cellsBuffer, 0, STUSB_RXBUF_SIZE);
	for( i=1; i<=cell_qty; i++)
	{
		if(i == cell_qty)
			pti = 2;
			
		cell[0] = (DEFAULT_GFC << 4) | (dev->vpi >> 4);
		cell[1] = (dev->vpi << 4) | (dev->vci >> 12);
		cell[2] = (dev->vci >> 4);
		cell[3] = ((dev->vci & 0xf) << 4) | pti;
		cell[4] = DEFAULT_HEC;
		memcpy(cell + ATM_CELL_HEADER, ptr, ATM_CELL_DATA);
		
		cell += ATM_CELL_SIZE;
		ptr  += ATM_CELL_DATA;
	}
	
	sz = cell_qty * ATM_CELL_SIZE;
	atomic_add(&(dev->write_on), 1);
	status = usb_write(dev, cellsBuffer, sz, NULL, STUSB_TIMEOUT);
	atomic_add(&(dev->write_on), -1);
#if 0
	dump_data(cellsBuffer, sz);
#endif
	return( status );
}

